#! /bin/sh -efu
# $Id: portcvsweb.sh,v 1.3 2006/10/07 10:59:58 securedog Exp $
#
# usage: portcvsweb.sh pkgname
#	 portcvsweb.sh /usr/{src,ports}/path
#	 cd /usr/{src,ports}/path; portcvsweb.sh
#	 cd /usr/{src,ports}; portcvsweb.sh path/file

set_variables() {
	if [ -r /usr/local/etc/portutil.d/options ]; then
		. /usr/local/etc/portutil.d/options
	fi

	: ${PORTSDIR="/usr/ports"}
	: ${PKGSRCDIR="/usr/pkgsrc"}
	: ${SRCDIR="/usr/src"}
	: ${BROWSER="w3m:lynx:links:firefox:mozilla:netscape"}
	: ${TARGET_OS="FreeBSD"}

	FreeBSD_URI_BASE="http://www.FreeBSD.org/cgi/cvsweb.cgi/"
	NetBSD_URI_BASE="http://cvsweb.netbsd.org/bsdweb.cgi/"
	OpenBSD_URI_BASE="http://www.openbsd.org/cgi-bin/cvsweb/"
	SUBDIR_SRC=src

	case ${TARGET_OS} in
	FreeBSD)
		SUBDIR_PORTS="ports" ;;
	NetBSD)
		PORTSDIR=${PKGSRCDIR}
		SUBDIR_PORTS="pkgsrc" ;;
	OpenBSD)
		SUBDIR_PORTS="ports" ;;
	*)
		return 1 ;;
	esac

	eval CVSWEB_URI=\"\${${TARGET_OS}_URI_BASE}\"
}

set_www_user_agent(){
	local IFS prog

	IFS=':'
	for prog in ${BROWSER}; do
		if prog=$(which "${prog}"); then
			WWW_USER_AGENT=${prog}
			return 0
		fi
	done

	echo "No browser is available - please define BROWSER." >&2
	return 1
}

set_target() {
	local path realpath

	path=$1
	realpath=

	if [ -z "$1" ]; then
		case ${PWD}/ in
		${PORTSDIR}/*|${SRCDIR}/*)
			realpath=${PWD} ;;
		*)
			realpath=/ ;;
		esac
	else
		if [ -e "${path}" ]; then
			realpath=$(realpath "${path}")
		else
			set -- $(portutil ports_glob "${path}")
			[ -n "${1-}" ] && realpath=${PORTSDIR}/$1
		fi
	fi

	if [ -n "${realpath}" ]; then
		[ -d "${realpath}" ] && realpath="${realpath}/"

		case ${realpath} in
		${PORTSDIR}/*)
			TARGET=${SUBDIR_PORTS}${realpath#${PORTSDIR}} ;;
		${SRCDIR}/*)
			TARGET=${SUBDIR_SRC}${realpath#${SRCDIR}} ;;
		//)
			TARGET= ;;
		*)
			echo "File not found: ${path}" >&2
			return 1
		esac
	else
		echo "Port directory not found: ${path}" >&2
		return 1
	fi
}

main() {
	set_variables || return 0
	set_www_user_agent || return 0
	set_target ${1-""} || return 0

	exec ${WWW_USER_AGENT} ${CVSWEB_URI}${TARGET}
}

main ${1+"$@"}
