#! /bin/sh -efu
# $Id: portutil.sh,v 1.5 2006/10/08 13:44:24 securedog Exp $

usage() {
	echo -n "usage: ${0##*/} ${command:-"command"} "
	if [ -z "${command}" ]; then
		cat <<EOF
[options] [[localopts] pkgname ...]

Commands:
  info		-- Information for the ports
  build		-- Build the ports
  fetch		-- Fetch the distfiles or packages for the ports
  install	-- Install the ports
  deinstall	-- Deinstall the packages
  update	-- Update the packages
  version	-- Compare the version of installed package with the ports tree
  package	-- Make the packages from the installed packages
  clean		-- Remove the working directories
  distclean	-- Remove the distfiles that are not referenced by the ports
  config	-- Configure options for the ports
  rmconfig	-- Remove the configured options for the ports
EOF
	else
		if [ -n "${OPTIONS}" ]; then
			echo -n "[-$(echo ${OPTIONS} | sed 's|.:||g')] "
			match ${OPTIONS} x: && echo -n "[-x pkgname_glob] "
		fi
		echo -n "["
		if [ -n "${LOCALOPTS}" ]; then
			echo -n "[-$(echo ${LOCALOPTS} | sed 's|.:||g')] "
			match ${LOCALOPTS} m: && echo -n "[-m make_args] "
			match ${LOCALOPTS} M: && echo -n "[-M make_env] "
			match ${LOCALOPTS} o: && echo -n "[-o origin] "
		fi
		echo "pkgname ...]"
	fi

	exit ${1:-0}
}

is_yes()
{
	case ${1-} in
	[Yy][Ee][Ss]|[Yy]|[Tt][Rr][Uu][Ee])
		return 0 ;;
	*)	return 1 ;;
	esac
}

warn() {
	echo "** $@" >&2
}

info() {
	is_yes ${opt_quiet} || echo "--->  $@"
}

match() {
	[ $# -eq 2 ] || return 1

	case $1 in
	*$2*)	return 0 ;;
	*)	return 1 ;;
	esac
}

prompt_yesno() {
	local prompt default input

	prompt=${1-"OK?"}
	default=${2-"yes"}

	echo -n "${prompt} [${default}] " >&2
	read input

	if is_yes ${input:-"${default}"}; then
		return 0
	else
		return 1
	fi
}

prompt_yesno_if_interactive() {
	if is_yes ${opt_noexecute}; then
		return 1
	elif is_yes ${opt_interactive}; then
		prompt_yesno ${1+"$@"} || return 1
	fi
}

reverse_list()
{
	local _list
	_list=
	while [ $# -gt 0 ]; do
		_list="$1 ${_list}"
		shift
	done
	echo ${_list}
}

init_global_options() {
	opt_all=NO
	opt_backup=YES
	opt_keep_backup=NO
	opt_force=NO
	opt_interactive=NO
	opt_use_index=NO
	opt_noexecute=NO
	opt_use_package=NO
	opt_noconf=NO
	opt_beforeclean=NO
	opt_afterclean=YES
	opt_verbose=NO
	opt_excludes=
	opt_quiet=NO
	opt_preserve_libs=YES
	opt_ignore_moved=NO
}

init_local_options() {
	opt_depends=NO
	opt_required_by=NO
	opt_origin=
	opt_make_args=
	opt_make_env=
}

init_options() {
	TYPE=ports

	case ${1-} in
	info)			;;
	build)			OPTIONS=c ;;
	fetch)			OPTIONS=P ;;
	install)		OPTIONS=cCPwW ;;
	deinstall)		TYPE=pkg ;;
	update)			TYPE=pkg; OPTIONS=bcCPuwWz; LOCALOPTS=o: ;;
	version)		TYPE=pkg; LOCALOPTS=o: ;;
	package|distclean)	TYPE=pkg ;;
	clean)			;;
	config|rmconfig)	;;
	*)			return 1 ;;
	esac

	OPTIONS=hafiInqvx:${OPTIONS}
	LOCALOPTS=rRm:M:${LOCALOPTS}
}

init_variables() {
	: ${PORTSDIR="/usr/ports"}
	: ${PACKAGES="${PORTSDIR}/packages"}
	: ${PACKAGEROOT="ftp://ftp.FreeBSD.org"}
	: ${PACKAGESITE=""}
	: ${PKGREPOSITORY="${PACKAGES}/All"}
	: ${PKG_DBDIR="/var/db/pkg"}
	: ${PKG_TMPDIR=${TMPDIR:-"/var/tmp"}}
	: ${PKG_SUFX=".tbz"}
	: ${PKG_BACKUP_DIR=${PKGREPOSITORY}}
	: ${PKGCOMPATDIR="/usr/local/lib/compat/pkg"}
	: ${MAKE="make"}
	OPTIONS=
	LOCALOPTS=
	confdir="/usr/local/etc/portutil.d"
	export PORTSDIR PACKAGES PKGREPOSITORY PKG_DBDIR PKG_TMPDIR PKG_SUFX PKGCOMPATDIR
}

init_categories() {
	local - cat

	if [ -z "${CATEGORIES-}" ]; then
		set +f
		for cat in ${PORTSDIR}/*/; do
			cat=${cat#${PORTSDIR}/}
			case ${cat%/} in
			[mM]k|[tT]emplates|Tools|bootstrap|distfiles|doc|licenses|packages) ;;
			*) CATEGORIES=${CATEGORIES:+"${CATEGORIES} "}${cat%/} ;;
			esac
		done
	fi
	: ${LANGUAGE_SPECIFIC_CATEGORIES="ar:arabic de:german fr:french hu:hungarian iw:hebrew ja:japanese ko:korean pl:polish pt:portuguese ru:russian uk:ukrainian vi:vietnamese zh:chinese"}
}


set_port_name() {
	port_name=
	if is_yes ${opt_use_index}; then
		port_name=$(grep --mmap -m 1 -o "^[^|]*|[^|]*${1#${PORTSDIR}/}|" "${PORTSDIR}/INDEX")
		port_name="${port_name%%|*}"
	elif [ -e "$1/Makefile" ] && cd "$1"; then
		port_name=$(${MAKE} -VPKGNAME)
	else
		return 1
	fi

	if [ -z "${port_name}" ]; then
		return 1
	fi
}

set_pkg_name() {
	pkg_name=$1
}

set_pkg_pkgdir() {
	pkg_pkgdir=${PKG_DBDIR}/$1
}

set_pkg_origin() {
	local LINE
	if [ -r "${PKG_DBDIR}/$1/+CONTENTS" ]; then
		while read LINE; do
			case ${LINE} in
			@comment\ ORIGIN:*)
				pkg_origin="${LINE#@comment ORIGIN:}"
				break ;;
			[!@]*)	break ;;
			esac
		done < "${PKG_DBDIR}/$1/+CONTENTS"
	fi
}

set_pkg_portdir() {
	pkg_portdir=${PORTSDIR}/$1
}

set_pkg_version() {
	pkg_version=${pkg_name##*-}
}

set_pkg() {
	local port_name

	case $1 in
	*/*)
		pkg_origin=$1
		set_pkg_portdir "${pkg_origin}"
		set_port_name "${pkg_portdir}"
		set_pkg_name "${port_name}" ;;
	*)
		set_pkg_name "$1"
		set_pkg_origin "$1"
		set_pkg_portdir "${pkg_origin}" ;;
	esac

	set_pkg_pkgdir "${pkg_name}"
}

set_make_cmd() {
	make_cmd="${opt_make_env:+env ${opt_make_env} }${MAKE}${opt_make_args:+ ${opt_make_args}}"
}

set_system_info() {
	: ${ARCH=$(uname -m)}
	: ${OS_VERSION=$(uname -r)}
	: ${OS_REVISION=${OS_VERSION%%-*}}
	: ${OS_MAJOR=${OS_REVISION%%.*}}
}

correct_pkg() {
	if [ -z "${pkg_origin}" ]; then
		if [ -d "${pkg_pkgdir}" ]; then
			warn "'${pkg_name}' has no origin recorded."
		else
			warn "'${pkg_name}' is not installed. (broken dependencies?)"
		fi
		return 1
	elif [ ! -d "${pkg_portdir}" ]; then
		warn "${pkg_name} (${pkg_origin}): perhaps moved or obsoleted."
		trace_moved_port "${pkg_origin}" || return 1
	fi
}

parse_options () {
	local OPTS OPT OPTARG OPTIND

	OPTS=$1; shift
	while getopts ${OPTS} OPT; do
		case ${OPT} in
		a)	opt_all=YES ;;
		b)	opt_keep_backup=YES ;;
		c)	opt_beforeclean=YES ;;
		C)	opt_afterclean=YES ;;
		h)	usage ;;
		f)	opt_force=YES ;;
		i)	opt_interactive=YES ;;
		I)	opt_use_index=YES ;;
		n)	opt_noexecute=YES ;;
		P)	opt_use_package=YES ;;
		q)	opt_noconf=YES ;;
		r)	opt_required_by=YES ;;
		R)	opt_depends=YES ;;
		u)	opt_preserve_libs=NO ;;
		v)	opt_verbose=YES ;;
		w)	opt_beforeclean=NO ;;
		W)	opt_afterclean=NO ;;
		z)	opt_backup=NO ;;
		x)	opt_excludes="${opt_excludes} ${OPTARG}" ;;
		m)	opt_make_args="${opt_make_args} ${OPTARG}" ;;
		M)	opt_make_env="${opt_make_args} ${OPTARG}" ;;
		o)	opt_origin="${OPTARG}" ;;
		\?)	usage 1 ;;
		esac
	done

	OPTC=$((${OPTIND}-1))
}

load_config() {
	[ -d "${confdir}" ] && [ -r "${confdir}" ] || opt_noconf=YES

	if ! is_yes ${opt_noconf}; then
		if [ -r "${confdir}/options" ]; then
			is_yes ${opt_verbose} && info "Sourcing ${confdir}/options"
			. "${confdir}/options"
		fi
	fi

	set_make_cmd

}

run_command() {
	local glob pkg_name pkg_origin pkg_portdir pkg_pkgdir

	glob="${TYPE}_glob"

	is_yes ${opt_required_by} && glob="${glob} -r"

	case ${command} in
	clean|config|rmconfig) ;;
	*)	is_yes ${opt_depends} && glob="${glob} -R" ;;
	esac

	set -- $(${glob} "${pkgname}")

	case ${command} in
	deinstall)
		if [ $# -gt 1 ]; then
			set -- $(pkg_sort "$@")
			set -- $(reverse_list "$@")
		fi ;;
	update)
		if [ $# -gt 1 ]; then
			if is_yes ${opt_use_index} && ! is_yes ${opt_force}; then
				set -- $(pre_parse_updates "$@")
				if [ $# -eq 0 ]; then
					warn "No need to upgrade packages"
					return 0
				fi
			fi
			set -- $(pkg_sort "$@")
		fi ;;
	esac

	if [ $# -eq 0 ]; then
		warn "No such installed package or ports: ${pkgname}"
		return 0
	fi

	while [ $# -gt 0 ]; do
		set_pkg "$1"; shift

		if [ -n "${opt_origin}" ] && pkg_match "${pkgname}"; then
			pkg_origin=${opt_origin}
			set_pkg_portdir "${pkg_origin}"
			opt_origin=
		fi

		correct_pkg || continue

		if [ -n "${opt_excludes}" ] && pkg_match ${opt_excludes}; then
			is_yes ${opt_verbose} && info "Skipping '${pkg_origin}' because it is excluded by user"
			continue
		fi

		set_make_cmd

		if ! run_make_${command}; then
			warn "Fix the problem and try again."
			echo
		fi
	done
}

pkg_match() {
	local pkg

	for pkg in ${1+"$@"}; do
		case "|${pkg_name%-[0-9]*[0-9]}|${pkg_origin}|" in
		"|${pkg%-[0-9]*[0-9]}|")
			return 0 ;;
		esac
	done
	return 1
}

extract_categories() {
	local pattern category

	pattern=$1
	if [ "${pattern}" = '*' ]; then
		categories="${CATEGORIES}"
	else
		for category in ${CATEGORIES}; do
			case ${category} in
			${pattern}) categories="${categories-} ${category}" ;;
			esac
		done
	fi
}

ports_glob() {
	local - OPTC pattern name portdir categories category opt_depends opt_required_by

	init_categories
	opt_depends=NO
	opt_required_by=NO
	parse_options rR ${1+"$@"}
	shift ${OPTC}

	set +f
	while [ $# -gt 0 ]; do
		pattern=${1#${PORTSDIR}/}; shift

		case ${pattern} in
		/*|*/|*/*/*|*./*|'') continue ;;
		*/*)
			name="${pattern#*/}"
			categories="${pattern%/*}" ;;
		??-*)
			name="${pattern}"
			categories='*'

			for category in ${LANGUAGE_SPECIFIC_CATEGORIES}; do
				if [ "${pattern%%-*}" = "${category%:*}" ]; then
					name="${pattern#*-}"
					categories="${category#*:}"
					break
				fi
			done ;;
		*)
			name="${pattern}"
			categories='*' ;;
		esac

		extract_categories "${categories}"

		for category in ${categories}; do
			for portdir in ${PORTSDIR}/${category}/${name}/; do
				if [ -d "${portdir}" ]; then
					portdir=${portdir%/}
					is_yes ${opt_depends} && ports_depends "${portdir}"
					echo "${portdir#${PORTSDIR}/}"
				fi
			done
		done
	done
}

ports_depends() {
	local checked
	checked=
	ports_depends_recursive "$1"
}

ports_depends_recursive() {
	local depend dep_portdir

	cd $1 || return 1
	for depend in $(${make_cmd} -VRUN_DEPENDS -VLIB_DEPENDS); do
		dep_portdir=${depend#*:}
		if ! match "${checked}" "|${dep_portdir}|"; then
			checked="${checked:-|}${dep_portdir}|"
			ports_depends_recursive "${dep_portdir}"
			echo "${dep_portdir#${PORTSDIR}/}"
		fi
	done
}

pkg_glob() {
	local - OPTC pattern contents opt_depends opt_required_by

	opt_depends=NO
	opt_required_by=NO
	parse_options rR ${1+"$@"}
	shift ${OPTC}

	set +f
	while [ $# -gt 0 ]; do
		pattern=${1#${PKG_DBDIR}/}; shift

		case ${pattern} in
		''|*[\<\>]*)	continue ;;
		*-[0-9]*|*\**)	;;
		*)		pattern="${pattern}-[0-9]*" ;;
		esac

		for contents in ${PKG_DBDIR}/${pattern}/+CONTENTS; do
			if [ -e "${contents}" ]; then
				name=${contents%/+CONTENTS}

				is_yes ${opt_depends} && pkg_depends "${name##*/}"
				echo "${name##*/}"
				is_yes ${opt_required_by} && pkg_required_by "${name#*/}"
			fi
		done
	done

	return 0
}

pkg_depends() {
	local dep
	if [ -r "${PKG_DBDIR}/$1/+CONTENTS" ]; then
		while read dep; do
			case ${dep} in
			@pkgdep\ *)
				echo "${dep#@pkgdep }" ;;
			[!@]*)	break ;;
			esac
		done < "${PKG_DBDIR}/$1/+CONTENTS"
	fi
}

pkg_required_by() {
	local req
	if [ -r "${PKG_DBDIR}/$1/+REQUIRED_BY" ]; then
		while read req; do
			echo "${req}"
		done < "${PKG_DBDIR}/$1/+REQUIRED_BY"
	fi
}

pkg_sort() {
	local leaves checked pkg_node pkg

	leaves=
	checked=
	pkg_node=" $@ "

	while [ $# -gt 0 ]; do
		case ${checked} in
		*" $1 "*)
			shift; continue ;;
		esac

		if [ ! -s "${PKG_DBDIR}/$1/+REQUIRED_BY" ]; then
			leaves="${leaves:- } $1 "
			checked="${checked:- } $1 "
			shift; continue
		fi

		for pkg in $(pkg_depends "$1") "$1"; do
			case ${checked} in
			*" ${pkg} "*) ;;
			*)	case ${pkg_node} in
				*" ${pkg} "*)
					checked="${checked:- }${pkg} "
					echo "${pkg}" ;;
				esac ;;
			esac
		done
		shift
	done

	for pkg in ${leaves}; do
		echo "${pkg}"
	done
}

pre_parse_updates() {
	local all name

	all=" $@ "
	for name in $(pkg_version -qvIl\< "${PORTSDIR}/INDEX"); do
		case ${all} in
		*" ${name%%  *} "*)
			echo "${name%%  *}" ;;
		esac
	done
}

try() {
	local exit

	if "$@"; then
		is_yes ${opt_verbose} && warn "Command successful: $@"
		return 0
	else
		exit=$?
		warn "Command failed (exit code ${exit}): $@"
		return ${exit}
	fi
}

run_make_info() {
	local masterdir dep descr distinfo

	descr="${pkg_portdir}/pkg-descr"
	distinfo="${pkg_portdir}/distinfo"

	info "Information for '${pkg_name}' (${pkg_origin})"
	prompt_yesno_if_interactive || return 0

	try cd ${pkg_portdir} || return 1
	echo
	echo "Comment:"
	${make_cmd} -VCOMMENT
	echo

	if [ -r "${descr}" ]; then
		echo "Description:"
		cat "${descr}"
		echo
	fi

	is_yes ${opt_verbose} || return 0

	echo "Depends on:"
	for dep in $(ports_depends "${pkg_portdir}"); do
		set_port_name "${PORTSDIR}/${dep}"
		echo "${port_name} (${dep})"
	done
	echo

	if [ -r "${distinfo}" ] || {
		distinfo=$(${make_cmd} -VMD5_FILE)
		[ -r "${distinfo}" ]
	}; then
		echo "Distfile(s):"
		sed -n 's,^S[Ii][Zz][Ee] (\(.*\)) = \(.*\),\1 (\2 bytes),p' "${distinfo}"
		echo
	fi
}

run_make_build() {
	if [ -n "${opt_make_args}" ]; then
		info "Building '${pkg_portdir}' with make flags: ${opt_make_args}"
	else
		info "Building '${pkg_portdir}'"
	fi
	prompt_yesno_if_interactive || return 0

	try cd ${pkg_portdir} || return 1
	if is_yes ${opt_beforeclean}; then
		opt_depends=YES run_make_clean || return 1
	fi

	try ${make_cmd} || return 1
}

run_make_fetch() {
	local fetch_cmd fetch_args

	if is_yes ${opt_use_package}; then
		info "Fetching the package for '${pkg_name}' (${pkg_origin})"
	else
		info "Fetching the distfile(s) for '${pkg_name}' (${pkg_origin})"
	fi
	prompt_yesno_if_interactive || return 0

	if is_yes ${opt_use_package}; then
		fetch_cmd="fetch"
		fetch_args="-ApRr"
		is_yes ${opt_verbose} && fetch_args="${fetch_args} -v"

		if [ -e "${PKGREPOSITORY}/${pkg_name}${PKG_SUFX}" ]; then
			info "Package '${pkg_name}${PKG_SUFX}' is already exists"
			return 0
		elif [ ! -w "${PKGREPOSITORY}" ]; then
			warn "${PKGREPOSITORY} is not writable."
			return 1
		else
			try mkdir -p "${PKGREPOSITORY}" || return 1
			try ${fetch_cmd} ${fetch_args} -o "${PKGREPOSITORY}" $(remote_pkg_uri "${pkg_name}") || return 1
		fi
	else
		try cd ${pkg_portdir} || return 1
		try ${make_cmd} checksum || return 1
	fi
}

remote_pkg_uri() {
	local name pkg_branch subdir

	name=$1
	if [ -n "${PACKAGESITE}" ]; then
		echo "${PACKAGESITE%/}/${name}${PKG_SUFX}"
		return 0
	fi

	set_system_info
	case ${OS_VERSION} in
	*-CURRENT)	pkg_branch="${OS_MAJOR}-current" ;;
	*-RELEASE*)	pkg_branch="${OS_REVISION}-release" ;;
	*)		pkg_branch="${OS_MAJOR}-stable" ;;
	esac

	case ${name} in
	*-[0-9]*[0-9]) subdir="All" ;;
	*) subdir="Latest" ;;
	esac

	echo "${PACKAGEROOT%/}/pub/FreeBSD/ports/${ARCH}/packages-${pkg_branch}/${subdir}/${name}${PKG_SUFX}"
}

run_make_install() {
	if ! is_yes ${opt_force} && [ -d "${PKG_DBDIR}/${pkg_name}" ]; then
		warn "${pkg_name} is already installed. (specify -f to force)"
		return 0
	fi

	if is_yes ${opt_use_package}; then
		info "Installing '${pkg_name}' from a package"
	else
		info "Installing '${pkg_name}' from a port (${pkg_origin})"
	fi
	prompt_yesno_if_interactive || return 0

	opt_interactive=NO
	if is_yes ${opt_use_package}; then
		run_make_fetch || return 1
	else
		run_make_build || return 1
	fi

	install_port || return 1
}

install_port() {
	local install_args

	install_args=
	if is_yes ${opt_use_package}; then
		is_yes ${opt_force} && install_args="-f"
		is_yes ${opt_verbose} && install_args="${install_args} -v"

		try pkg_add ${install_args} "${PKGREPOSITORY}/${pkg_name}${PKG_SUFX}" || return 1
	else
		is_yes ${opt_force} && install_args="-DFORCE_PKG_REGISTER"

		try cd ${pkg_portdir} || return 1
		try ${make_cmd} ${install_args} install || return 1

		if is_yes ${opt_afterclean}; then
			opt_depends=YES run_make_clean || return 1
		fi
	fi
}

run_make_deinstall() {
	local deinstall_args

	info "Deinstalling '${pkg_name}'"
	prompt_yesno_if_interactive || return 0

	deinstall_args=
	is_yes ${opt_force} && deinstall_args="-f"
	is_yes ${opt_verbose} && deinstall_args="${deinstall_args} -v"

	if [ ! -w "${PKG_DBDIR}" ]; then
		warn "You do not own ${PKG_DBDIR}."
		return 1
	fi

	try pkg_delete ${deinstall_args} "${pkg_name}" || return 1
}

run_make_update() {
	local UPDATE_TMPDIR old_package old_required_by preserved_files port_name status

	set_port_name "${pkg_portdir}" || return 1
	status=$(compare_version "${pkg_name##*-}" "${port_name##*-}") || return 1

	if [ ${status} = '<' ]; then
		info "Upgrading '${pkg_name}' to '${port_name}' (${pkg_origin})"
	elif ! is_yes ${opt_force}; then
		is_yes ${opt_verbose} && \
		warn "No need to upgrade '${pkg_name}'. (>= ${port_name}). (specify -f to force)"
		return 0
	elif [ ${status} = '=' ]; then
		info "Reinstalling '${pkg_name}' (${pkg_origin})"
	elif [ ${status} = '>' ]; then
		info "Downgrading '${pkg_name}' to '${port_name}' (${pkg_origin})"
	fi
	prompt_yesno_if_interactive || return 0

	opt_interactive=NO
	if is_yes ${opt_use_package}; then 
		pkg_name=${port_name} run_make_fetch || return 1
	else
		pkg_name=${port_name} run_make_build || return 1
	fi

	UPDATE_TMPDIR=$(try mktemp -d "${PKG_TMPDIR}/${pkg_name}.XXXXXX") || return 1
	old_package=${UPDATE_TMPDIR}/${pkg_name}${PKG_SUFX}
	old_required_by=${UPDATE_TMPDIR}/+REQUIRED_BY

	if is_yes ${opt_backup}; then
		info "Backing up the old version"
		opt_quiet=YES PKGREPOSITORY=${UPDATE_TMPDIR} run_make_package || return 1
	fi

	if [ -e "${pkg_pkgdir}/+REQUIRED_BY" ]; then
		info "Backing up the +REQUIRED_BY file"
		try cp -f "${pkg_pkgdir}/+REQUIRED_BY" "${old_required_by}" || return 1
	fi

	if is_yes ${opt_preserve_libs}; then
		preserve_shlibs || return 1
	fi

	if opt_force=YES run_make_deinstall; then
		if is_yes ${opt_use_package}; then
			info "Installing the new version via the package"
		else
			info "Installing the new version via the port"
		fi

		if install_port; then
			# Set new PKGNAME via the Makefile
			status=done
			opt_use_index=NO set_port_name "${pkg_portdir}"
			set_pkg_pkgdir "${port_name}"
		else
			warn "Install failed."

			if is_yes ${opt_backup}; then
				info "Restoring the old version"
				try pkg_add -f "${old_package}" || return 1
			fi
		fi
	else
		warn "Deinstall failed."
	fi

	if [ -e "${old_required_by}" ] && [ -d "${pkg_pkgdir}" ] && \
		[ ! -e "${pkg_pkgdir}/+REQUIRED_BY" ]; then
		info "Restoring the +REQUIRED_BY file"
		try mv -f "${old_required_by}" "${pkg_pkgdir}" || return 1
	fi

	if is_yes ${opt_backup} && is_yes ${opt_keep_backup}; then
		info "Keeping old package in '${PKG_BACKUP_DIR}'"
		try mkdir -p "${PKG_BACKUP_DIR}" || return 1
		try mv -f "${old_package}" "${PKG_BACKUP_DIR}" || return 1
	fi

	if is_yes ${opt_preserve_libs}; then
		clean_shlibs || warn "Couldn't remove the preserved shared libraries."
	fi

	info "Removing temporary files and directories"
	try rm -rf "${UPDATE_TMPDIR}" || warn "Couldn't remove '${UPDATE_TMPDIR}'."

	if [ ${status} != done ]; then
		warn "Failed to upgrade the ${pkg_name}."
		return 1
	elif [ "${pkg_name}" != "${port_name}" ]; then
		info "Updating the pkgdb"
		try pkgdb_fix -s "/${pkg_name%-*}-[^-]*/${port_name}/" -r "${port_name}" || return 1
	fi
}

trace_moved_port() {
	local IFS LINE moved

	if [ ! -e "${PORTSDIR}/MOVED" ]; then
		warn "'${PORTSDIR}/MOVED' does not exist."
		return 1
	fi

	pkg_origin=$1
	moved=
	IFS='
'
	for LINE in $(grep -v '^#' "${PORTSDIR}/MOVED"); do
		case ${LINE} in
		${pkg_origin}\|\|*)
			warn "'${pkg_origin}' has been removed from ports tree:"
			warn "${LINE##**|}"
			return 1 ;;
		${pkg_origin}\|*\|*)
			moved=${LINE#*|}
			warn "'${pkg_origin}' has been moved to '${moved%%|*}'."
			pkg_origin=${moved%%|*} ;;
		esac
	done

	if [ -n "${moved}" ]; then
		set_pkg_portdir "${pkg_origin}"
	else
		warn "Port directory not found: ${pkg_origin}"
		return 1
	fi
}

preserve_shlibs() {
	local file

	preserved_files=
	for file in $(pkg_info -qL "${pkg_name}"); do
		case ${file} in
		*.so.[0-9]*)
			case $(file -b "${file}") in
			*ELF*shared\ object*)
				preserved_files="${preserved_files} ${file}"
			esac ;;
		esac
	done

	if [ -n "${preserved_files}" ]; then
		info "Preserving the shared libraries"
		try mkdir -p "${PKGCOMPATDIR}" || return 1
		try cp -f ${preserved_files} "${PKGCOMPATDIR}" || return 1
		try ldconfig -m "${PKGCOMPATDIR}" || return 1
	fi
}

clean_shlibs() {
	local delete_files file dest

	[ -n "${preserved_files}" ] || return 0

	info "Cleaning the preserved shared libraries"

	delete_files=
	for file in ${preserved_files}; do
		dest=${PKGCOMPATDIR}/${file##*/}
		if [ -e "${file}" ]; then
			delete_files="${delete_files} ${dest}"
		else
			info "Keeping ${file} as ${dest}"
		fi
	done

	if [ -n "${delete_files}" ]; then
		try rm -f ${delete_files} || return 1
		try ldconfig -m "${PKGCOMPATDIR}" || return 1
	fi
}

run_make_version() {
	local port_name port_version status

	set_port_name "${pkg_portdir}" || return 1
	port_version=${port_name##*-}
	status=$(compare_version "${pkg_name##*-}" "${port_version}") || return 1

	is_yes ${opt_verbose} || pkg_name=${pkg_name%-*}
	printf "%-34s  %c" "${pkg_name}" "${status}"

	if is_yes ${opt_verbose}; then
		echo -n "   "
		case ${status} in
		=)	echo -n "up-to-date with port" ;;
		\<)	echo -n "needs updating (port has ${port_version})" ;;
		\>)	echo -n "succeeds port (port has ${port_version})" ;;
		\!)	echo -n "comparison failed" ;;
		esac
	fi
	echo
}

compare_version() {
	if [ "$1" = "$2" ]; then
		echo '='
	else
		pkg_version -t "$1" "$2" || return 1
	fi
}

run_make_package() {
	local file

	file=${PKGREPOSITORY}/${pkg_name}${PKG_SUFX}
	if ! is_yes ${opt_force} && [ -e "${file}" ]; then
		info "Package '${file##*/}' is already exists (specify -f to force)"
		return 0
	fi

	info "Creating package for '${pkg_name}' (${pkg_origin})"
	prompt_yesno_if_interactive || return 0

	try mkdir -p "${file%/*}" || return 1
	if [ ! -w "${file%/*}" ]; then
		warn "${file%/*} is not writable."
		return 1
	fi

	try pkg_create -b "${pkg_name}" "${file}" || return 1
	info "Package created in '${file%/*}'"
}

run_make_clean() {
	info "Cleaning for ${pkg_name}"
	prompt_yesno_if_interactive || return 0

	try cd ${pkg_portdir} || return 1
	if is_yes ${opt_depends}; then
		try ${make_cmd} clean
	else
		try ${make_cmd} NOCLEANDEPENDS=YES clean
	fi
}

run_make_distclean() {
	local - distfiles file pattern match

	info "Distribution cleaning for '${pkg_name}' (${pkg_origin})"
	try cd ${pkg_portdir} || return 1

	distfiles=$(${make_cmd} -V_DISTFILES)
	try cd $(${make_cmd} -V_DISTDIR) || return 1

	set +o noglob
	for file in ${distfiles}; do
		pattern=$(echo "${file}" | \
		sed -E	-e 's/[[:digit:]][[:digit:].]{2,}/[0-9]*/g' \
			-e 's/[[:<:]](alpha|beta|pre|rc)[[:digit:]]*/*/g')

		is_yes ${opt_verbose} && info "Searching ${PWD}/${pattern}"

		for match in ${pattern}; do
			case ${match} in
			${file}|*\**) break ;;
			esac 

			if is_yes ${opt_noexecute}; then
				info "Skipping ${match}"
			elif prompt_yesno_if_interactive "Delete ${match}?"; then
				info "Deleting ${match}"
				rm -f "${match}"
			fi
		done
	done
}

run_make_config() {
	info "Configuration for '${pkg_name}' (${pkg_origin})"
	prompt_yesno_if_interactive || return 0

	try cd ${pkg_portdir} || return 1
	if is_yes ${opt_depends}; then
		try ${make_cmd} config-recursive
	else
		try ${make_cmd} config
	fi
}

run_make_rmconfig() {
	info "Removing configured options for '${pkg_name}' (${pkg_origin})"
	prompt_yesno_if_interactive || return 0

	try cd ${pkg_portdir} || return 1
	if is_yes ${opt_depends}; then
		try ${make_cmd} rmconfig-recursive
	else
		try ${make_cmd} rmconfig
	fi
}

main() {
	local command pkgname

	command=${1-}; shift || usage

	init_variables
	init_global_options
	init_local_options

	case ${command} in
	pkg_glob|pkg_sort|ports_glob)
		load_config
		${command} ${1+"$@"}
		exit 0 ;;
	esac

	init_options ${command} || command= usage
	parse_options ${OPTIONS}${LOCALOPTS} ${1+"$@"}
	shift ${OPTC}
	load_config

	if is_yes ${opt_use_index} && [ ! -e "${PORTSDIR}/INDEX" ]; then
		warn "'${PORTSDIR}/INDEX' does not exist."
		warn "Couldn't use -I option."
		exit 1
	fi

	if is_yes ${opt_all}; then
		opt_depends=NO
		opt_required_by=NO
		set -- '*'
	elif [ $# -eq 0 ]; then
		case ${PWD} in
		${PORTSDIR}/*/*) set -- "${PWD#${PORTSDIR}/}" ;;
		*) usage ;;
		esac
	fi

	while [ $# -gt 0 ]; do
		case $1 in
		-*)
			parse_options ${LOCALOPTS} "$@"
			shift ${OPTC} ;;
		esac

		pkgname=${1-}; shift || break
		run_command

		init_local_options
	done

	exit 0
}

IFS=' 
'
main ${1+"$@"}
